#!/usr/bin/ruby

#
# ChordExplorer: simple application to explore different variants of
# playing guitar chords
#

# Copyright (C) 2002 Alexander Tsvyashchenko
#
# This program is free software; you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program; if not,
# write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

require 'tk'
require 'tk/scrollbox'

class Note
  @@intervals = {'0' => 0, '1' => 0,
                 'b2' => 1,
		 '2' => 2,
		 'b3' => 3,
		 '3' => 4, '#3' => 4,
		 '4' => 5,
		 'b5' => 6,
		 '5' => 7,
		 '#5' => 8, 'b6' => 8,
		 '6' => 9, '#6' => 9,
		 'b7' => 10,
		 '7' => 11, '#7' => 11,
		 '8' => 12,
		 'b9' => 13,
		 '9' => 14,
		 '#9' => 15, 'b10' => 15,
		 '10' => 16, '#10' => 16, 'b11' => 16,
		 '11' => 17,
		 '#11' => 18,
		 '12' => 19,
		 'b13' => 20,
		 '13' => 21}
		 
  def initialize(note)
    @oct = 1
    if note.upcase =~ /([A-G][B#]?)(\d+)?/
      @note = $1
      @oct = $2.to_i if $2 != nil
    else
      raise 'Unknown note'
    end
    canonify
  end
  
  def ==(note)
    self.to_s == note.to_s
  end
  
  def +(int)
    @@intervals[int.downcase].times {self.succ}
    self
  end

  def next_note(i)
    i.times {self.succ}
    self
  end

  def prev_note(i)
    i.times {self.pred}
    self
  end

  def -(int)
    @@intervals[int.downcase].times {self.pred}
    self
  end
  
  def succ
    if @note.length == 1
      @note += '#'
    elsif @note[1].chr == 'B'
      @note = @note[0].chr
    elsif @note[0].chr == 'G'
      @note = 'A'
    else
      raise "Internal error" if @note[1].chr != '#'
      @note = (@note[0]+1).chr
    end
    canonify
    self
  end

  def pred
    if @note.length == 1
      @note += 'B'
    elsif @note[1] == '#'
      @note = @note[0].chr
    elsif @note[0].chr == 'A'
      @note = 'G'
    else
      raise "Internal error" if @note[1].chr != 'B'
      @note = (@noe[0]-1).chr
    end
    canonify
    self
  end

  def canonify
    case @note
      when 'E#'
        @note = 'F'
      when 'FB'
        @note = 'E'
      when 'B#'
        @note = 'C'
        @oct += 1
      when 'CB'
        @note = 'B'
        @oct -= 1
    end
  end
  
  def to_s
    return @note + @oct.to_s if @note.length == 1
    return @note[0].chr + 'b' + @oct.to_s if @note[1].chr == 'B'
    raise "Internal error" if @note[1].chr != '#'
    return 'Ab' + @oct.to_s if @note == 'G#'
    return (@note[0]+1).chr+'b'+@oct.to_s
  end
  
  def skip_oct
    self.to_s[0...-1]
  end
end

class Chord
  attr_reader :root, :name, :hls
  def initialize(root, name, hls)
    @root = root
    @name = name
    @hls = hls
  end
end

class ChordStructure
  attr_reader :name, :intervals, :alias_name
  def initialize(name, intervals, alias_name = nil)
    @name = name
    @intervals = intervals
    @alias_name = alias_name
  end
  def intervals_to_s
    str = ''
    @intervals.each {|i| str += i + ' '}
    str.chomp
  end
end

class ChordMain
  @@notes = %w{C Db D Eb E F Gb G Ab A Bb B}
  @@tuning = %w{E1 A1 D2 G2 B2 E3}
  def initialize()
    sec_type = nil
    @chords = Array.new
    @chords_structs = Array.new
    @show_chord_erasable = Array.new
    @show_notes_erasable = Array.new
    @chords_names = Array.new
    
    @strings_dist = 25
    @frets_dist =  40
    @frets_count = 20
    @string_color = '#404040'
    @fret_color = '#A06000'
    @strings_count = 6
    @shift_sl_x = 20
    @shift_y = 40
    @shift_strings_x = 65
    @strings_count = 6
    @finger_shift = 8
    @finger_color_bk = '#FFFFFF'
    @finger_color_fg = '#000000'
    @int_strict = '#00FF00'
    @int_non_strict = '#00FFFF'
    @tick_len = 10
    @labels=[3,5,7,12,14]
    @chord_name = ''
    @fret_max_dist = 3
     
#    IO.foreach(ENV['HOME'] + '/' + '.chordexplorer')  { |line|
    IO.foreach('.chordexplorer')  { |line|
      if line =~ /^\[(.+)\]/
	case sec_name = $1.downcase
	  when 'chords structures', 'chords', 'params'
	    sec_type = sec_name
	    next
	end
      else
	case sec_type
	  when 'chords structures'
	    if line =~ /.+=.+/
	      (name, ints) = line.chomp.split(/\s*=\s*/)
              puts name
              puts ints
	      name.chomp(' ')
	      struct = ints.scan(/[b#]?\d+/)
	      @chords_structs.push(ChordStructure.new(name, struct))
	      next
	    elsif line =~ /([^\s]+)\s+alias\s+([^\s]+)/
	      @chords_structs.push(ChordStructure.new($1,nil,$2))
	      next
	    end
	  when 'chords'
	    if line =~ /(\w+)\s+(.+)\s*=\s*(.+)/
	      if !(@@notes.find {|note| note == $1})
		puts "Warning: chord root #{$1.dump} is not a valid note name"
	      end
	      root = $1
	      name = $2
	      hls = $3.scan(/\d+/)
	      @chords.push(Chord.new(root, name, hls))
	      next
	    end
	  when 'params'
	    if line.downcase =~ /(@strings_dist|@frets_dist|@frets_count)\s*=\s*(\d+)/
              eval "#{$1}=#{$2}"
	      next
	    elsif line.downcase =~ /(@fret_color|@string_color)\s*=\s*(#?\w+)/
	      eval "#{$1}='#{$2}'"
	      next
	    end
	end
      end
      
      if line.chomp != '' && (line !~ /#\.*/)
        puts "Wrong config file: cannot parse #{line.dump}"
        exit 1
      end
    }
    
    @chords.each {|chord| puts "Warning: there is no chord structure called #{chord.name.dump}" if !(@chords_structs.find {|ch| ch.name == chord.name})}
    
    @root_win = TkRoot.new { title 'Chord Explorer' }
    

    main_frame = TkFrame.new(@root_win) {pack('side' => 'top')}
    frame_left = TkFrame.new(main_frame) {pack('side' => 'left')}
    frame_right = TkFrame.new(main_frame) {pack('side' => 'right')}

    frame = TkFrame.new(frame_left) {pack('side' => 'top')}
    TkLabel.new(frame) {text  'Root note:'; pack('side' => 'left')}
    @root_note = TkVariable.new
    mb = TkOptionMenubutton.new(frame, @root_note, *@@notes) {pack('side' => 'right')}
    @@notes.length.times { |i|
      mb.menu.entryconfigure(i, 'command' => proc {recalc_all()})
    }
    
    frame = TkFrame.new(frame_left) {pack('side' => 'top')}
    @show_aliases = TkVariable.new
    TkLabel.new(frame, 'text' => 'Show aliases:').pack('side' => 'left')
    TkCheckButton.new(frame, 'variable' => @show_aliases, 'command' => proc {fill_chords_names_list(); show_chord()}).pack('side' => 'right')

    frame = TkFrame.new(frame_right) {pack('side' => 'top')}
    TkLabel.new(frame) {text  'Chord type:'; pack('side' => 'top')}
    @chords_names_list = TkScrollbox.new(frame) {pack('side' => 'bottom')}
    fill_chords_names_list()
    @chords_names_list.bind('<ListboxSelect>', proc {@chord_name = @chords_names[@chords_names_list.curselection()[0].to_i]; resolve_aliases(); recalc_all()})

    frame = TkFrame.new(frame_left) {pack('side' => 'top')}
    @show_chord = TkVariable.new(1)
    TkLabel.new(frame, 'text' => 'Show predefined chord if available:').pack('side' => 'left')
    TkCheckButton.new(frame, 'variable' => @show_chord, 'command' => proc {show_chord()}).pack('side' => 'right')
    
    frame = TkFrame.new(frame_left) {pack('side' => 'top')}
    @show_notes = TkVariable.new(1)
    TkLabel.new(frame, 'text' => 'Show notes of chord intervals:').pack('side' => 'left')
    TkCheckButton.new(frame, 'variable' => @show_notes, 'command' => proc {show_notes()}).pack('side' => 'right')

    frame = TkFrame.new(frame_left) {pack('side' => 'top')}
    @show_chord_variant = TkVariable.new
    TkLabel.new(frame, 'text' => 'Show given variant of chord:').pack('side' => 'top')
    @scale = TkScale.new(frame, 'variable' => @show_chord_variant, 'orient' => 'horizontal', 'from' => 1, 'to' => 1, 'command' => proc {show_chord()}).pack('side' => 'bottom')
    
    frame = TkFrame.new(frame_left) {pack('side' => 'top')}
    @show_root_variant = TkVariable.new
    TkLabel.new(frame, 'text' => 'Show given variant of root note:').pack('side' => 'top')
    @scale_root = TkScale.new(frame, 'variable' => @show_root_variant, 'orient' => 'horizontal', 'from' => 1, 'to' => 1, 'command' => proc {recalc_notes()}).pack('side' => 'bottom')

    @chord_info = TkLabel.new(@root_win, 'text' => '') {pack('side' => 'top')}

    frame = TkFrame.new(@root_win) {pack('side' => 'bottom', 'fill' => 'x')}
    @canvas = TkCanvas.new(frame, 'scrollregion' => [0, 0, @frets_count*@frets_dist + @shift_strings_x, 0]) {pack('side' => 'top', 'fill' => 'x')}
    @hsb = TkScrollbar.new(frame) { orient 'horizontal'; pack('side' => 'bottom', 'fill' => 'x')}
    @canvas.xscrollbar(@hsb)

    draw_fretboard()
    
    recalc_all()

    Tk.mainloop
  end

  def fill_chords_names_list()
    @chords_names_list.delete(0, 'end')
    @chords_names.clear()
    @chords_structs.each  {|ch|
    if @show_aliases.value == '1'  || ch.alias_name == nil
      str = ch.name
      str += ' --> '+ch.alias_name if ch.alias_name != nil
      @chords_names_list.insert('end', str)
      @chords_names.push(ch.name)
    end
    }
  end
  
  def draw_fretboard()
    @strings_count.times {|i|
      TkcText.new(@canvas, @shift_sl_x, i*@strings_dist+@shift_y, 'text' => @@tuning[@strings_count-i-1])
      
      TkcLine.new(@canvas, @shift_strings_x, i*@strings_dist + @shift_y, @shift_strings_x + @frets_count*@frets_dist, i*@strings_dist + @shift_y, 'fill' => @string_color)
    }
      
    @frets_count.times {|i|
      TkcLine.new(@canvas, i*@frets_dist + @shift_strings_x, @shift_y - @tick_len, i*@frets_dist + @shift_strings_x, (@strings_count - 1)*@strings_dist + @shift_y + @tick_len, 'fill' => @fret_color, 'width' => 2)
    }
    @labels.each { |i|
    TkcText.new(@canvas, i*@frets_dist + @shift_strings_x - @frets_dist/2, @shift_y+@strings_count*@strings_dist, 'text' => i.to_s)
    }
  end
  
  def resolve_aliases()
    chord = @chords_structs.find {|ch| ch.name == @chord_name}
    if chord.alias_name != nil
      chord_new = @chords_structs.find { |ch| ch.name == chord.alias_name}
       @chord_name = chord_new.name if chord_new != nil
    end
  end
  
  def recalc_notes()
    if @root_note.to_s != '' && @chord_name != ''
      variants = 0
      r_note = Note.new(@root_note.to_s).skip_oct
      
      @strings_count.times { |string|
        @frets_count.times { |fret|
          note = Note.new(@@tuning[string]).next_note(fret)
          if r_note == note.skip_oct
	    variants += 1
	    if variants == @show_root_variant.value.to_i
	      @d_r_note = note
	      @d_r_fret = fret
	    end
	  end
        }
      }
      if variants == 0
        @scale_root.configure('state' => 'disabled')
        @scale_root.configure('from' => 0)
      else
        @scale_root.configure('state' => 'normal')
        @scale_root.configure('from' => 1)
      end
      @scale_root.configure('to' => variants)
    
      ch = @chords_structs.find {|c| c.name == @chord_name}
      str = ''
      ch.intervals.each {|i| str += (Note.new(@d_r_note.to_s) + i).to_s + ' '}
      @chord_info.text("Chord: #{@root_note.to_s}#{ch.name}\nIntervals: #{ch.intervals_to_s}\nNotes: #{str}")
      
      show_notes()
    end
  end

  def recalc_chords()
    if @root_note.to_s != '' && @chord_name != ''
      variants = 0
      @chords.each { |chord|
        variants += 1 if chord.root == @root_note.to_s && chord.name == @chord_name.to_s
      }
      if variants == 0
        @scale.configure('state' => 'disabled')
        @scale.configure('from' => 0)
      else
        @scale.configure('state' => 'normal')
        @scale.configure('from' => 1)
      end
      @scale.configure('to' => variants)

      show_chord()
    end
  end
  
  def recalc_all()
    recalc_notes()
    recalc_chords()
  end
  
  def show_chord()
    @show_chord_erasable.each {|item| item.destroy}
    @show_chord_erasable.clear()
    if @show_chord == '1' && @root_note.to_s != '' && @chord_name != ''
      variant = 0 
      @chords.each { |chord|
        if chord.root == @root_note.to_s && chord.name == @chord_name.to_s
	  variant += 1
	  next if variant < @show_chord_variant.value.to_i
	  min_fret = 255
	  (chord.hls.length/2).times { |i|
	    digit = chord.hls[2*i].to_i
	    finger = chord.hls[2*i+1]
	    if digit != 0
	      x = (digit-1)*@frets_dist + @shift_strings_x - @frets_dist/2
	      y = i*@strings_dist + @shift_y
	      min_fret = (min_fret > digit)? digit: min_fret
	    
	      @show_chord_erasable.push(
	        TkcOval.new(@canvas, x - @finger_shift,y - @finger_shift, x + @finger_shift, y + @finger_shift, 'fill' => @finger_color_bk))
	    
	      @show_chord_erasable.push(
	        TkcText.new(@canvas, x, y, 'text' => finger, 'fill' => @finger_color_fg))
	    end
	  }
	  @show_chord_erasable.push(
	    TkcText.new(@canvas, (min_fret-1)*@frets_dist + @shift_strings_x - @frets_dist/2, @shift_y - 2*@tick_len, 'text' => (min_fret-1).to_s))
	  @canvas.xview('moveto', (min_fret>0? (min_fret-2).to_f: 0) / @frets_count)
	  break
        end # chord equals
      }
    end # show_chord == '1'
  end
  
  def show_notes()
    @show_notes_erasable.each {|item| item.destroy}
    @show_notes_erasable.clear()
    busy_notes = Hash.new(0)
    if @show_notes == '1' && @root_note.to_s != '' && @chord_name != ''
      ch = @chords_structs.find {|c| c.name == @chord_name}
      
      @strings_count.times { |string|
	@frets_count.times { |fret|
	  note = Note.new(@@tuning[string]).next_note(fret)
	  ch.intervals.length.times { |i|
	    if (Note.new(@root_note.to_s) + ch.intervals[i]).skip_oct == note.skip_oct
	      next if fret != 0 && (fret-@d_r_fret).abs > @fret_max_dist
	      # fl_strict = (Note.new(@root_note.to_s) + ch.intervals[i] == note)
	      fl_strict = (Note.new(@d_r_note.to_s) + ch.intervals[i] == note)
              x = fret*@frets_dist + @shift_strings_x - @frets_dist/2
	      y = (@strings_count - string - 1)*@strings_dist + @shift_y
	      busy_notes[[x,y]] += 1
              if busy_notes[[x,y]] > 1
		x += 5*(busy_notes[[x,y]]-1)
		y += 5*(busy_notes[[x,y]]-1)
	      end
	      @show_notes_erasable.push(
	        TkcRectangle.new(@canvas, x - @finger_shift*1.4, y - @finger_shift, x + @finger_shift*1.4, y + @finger_shift, 'fill' => fl_strict ? @int_strict : @int_non_strict, 'width' => fl_strict ? '2' : '1'))
	      @show_notes_erasable.push(
	        TkcText.new(@canvas, x, y, 'text' => ch.intervals[i], 'fill' => @finger_color_fg))
	    end
      	  }
	}
      }
      @show_notes_erasable.push(
	TkcText.new(@canvas, @d_r_fret*@frets_dist + @shift_strings_x - @frets_dist/2, @shift_y - 2*@tick_len, 'text' => @d_r_fret.to_s))
	  @canvas.xview('moveto', (@d_r_fret>1? (@d_r_fret-2).to_f: 0) / @frets_count)
    end
  end
end

ChordMain.new
    
